@@ -5,16 +5,22 @@ import enum
 from typing import Optional, List
 from pathlib import Path
 
-from . mi_api import MiApi
-from . mi_endpoints import mi_endpoint_get
+from . mi_api import MiApi, mi_api_execute
+from . mi_api import mi_meta
 
+from . mi_endpoints import mi_endpoint_get
 from . mi_endpoints import mi_endpoint_delete
 from . mi_endpoints import mi_endpoint_get
 from . mi_endpoints import mi_endpoint_list
 from . mi_endpoints import mi_endpoint_save
 
+from . mi_models import Endpoint, MiRecords
+from . mi_models import MiProgramMetadata
+from . mi_models import MiPrograms
+from . mi_models import MiTransactionMetadata
+from . mi_models import MiError
+
 from . mi_outputs import mi_output_generate_template
-from . mi_models import Endpoint
 
 
 Keys = enum.Enum(
@@ -50,34 +56,20 @@ def execute(endpoint: str, program: str, transaction: str, params: Optional[List
         typer.echo(f'Failed to parse extra parameters!\n {params}\n Has to have format KEY=VALUE')
         return
     
-    endpoint = mi_endpoint_get(endpoint)
-    
-    api = MiApi(endpoint.host, endpoint.port, endpoint.usr, endpoint.pwd)
-    records = api.execute(program, transaction, **params)
+    records = mi_api_execute(endpoint, program, transaction, **params)
 
-    print('\n')
-    print(
-        tabulate.tabulate(
-            [
-                record for record
-                in records.records
-            ], headers='keys'
+    if isinstance(records, MiRecords):
+        print('\n')
+        print(
+            tabulate.tabulate(
+                [
+                    record for record
+                    in records.records
+                ], headers='keys'
+            )
         )
-    )
-
-
-@app.command()
-def meta(endpoint: Names, program: Optional[str] = typer.Argument(None), transaction: Optional[str] = typer.Argument(None)):
-    ''' Show metadata by listing programs, program transaction and transaction details'''
-
-    if not program and not transaction:
-        print_program_list(endpoint.name)
-
-    elif program and not transaction:
-        print_program_meta(endpoint.name, program)
-
-    elif program and transaction:
-        print_transaction_meta(endpoint.name, program, transaction)
+    else:
+        print(records)
 
 
 @app.command()
@@ -90,9 +82,9 @@ def template(endpoint_name: Names, program: str, transaction: str, path: Optiona
         return
 
     endpoint = mi_endpoint_get(endpoint_name.value)
-    api = MiApi(endpoint.host, endpoint.port, endpoint.usr, endpoint.pwd)
     mi_output_generate_template(
-        api.transaction_meta(
+        mi_meta(
+            endpoint_name.value,
             program,
             transaction
         ),
@@ -103,7 +95,19 @@ def template(endpoint_name: Names, program: str, transaction: str, path: Optiona
 def endpoint(name: Optional[str] = typer.Argument(None), delete: Optional[bool] = None):
     ''' List or configure the API endpoints '''
     if not name:
-        print_endpoint_list()
+        print(
+            tabulate.tabulate(
+                [
+                    {
+                        key: val for (key, val) in endpoint.__dict__.items()
+                        if key != 'pwd'
+                    }
+                    for endpoint
+                    in mi_endpoint_list()
+                ],
+                headers='keys'
+            )
+        )
         return
 
     if delete:
@@ -150,84 +154,58 @@ def endpoint(name: Optional[str] = typer.Argument(None), delete: Optional[bool]
     mi_endpoint_save(endpoint)
     typer.echo(f'Saved {name}!')
 
+@app.command()
+def meta(endpoint: Names, program: Optional[str] = typer.Argument(None), transaction: Optional[str] = typer.Argument(None)):
+    metadata = mi_meta(endpoint.value, program, transaction)
 
-def print_transaction_meta(endpoint: str, program: str, transaction: str):
-    endpoint = mi_endpoint_get(endpoint)
+    if isinstance(metadata, MiPrograms):
+        for row in metadata.records:
+            typer.echo(row)
+
+    elif isinstance(metadata, MiProgramMetadata):
+        typer.echo(f'{metadata.program} - {metadata.description}\n')
     
-    api = MiApi(endpoint.host, endpoint.port, endpoint.usr, endpoint.pwd)
-    records = api.transaction_meta(program, transaction)
-
-    typer.echo(f'{records.program}/{records.transaction}')
-    typer.echo('\nInputs:')
-    print(
-        tabulate.tabulate(
-            [
-                input.__dict__ for input in
-                records.inputs
-            ],
-            headers='keys'
+        print(
+            tabulate.tabulate(
+                [
+                    {
+                        'transaction': transaction.transaction,
+                        'description': transaction.description
+                    }
+                    for transaction in
+                    metadata.transactions
+                ],
+                headers='keys'
+            )
         )
-    )   
+    elif isinstance(metadata, MiTransactionMetadata):
+        typer.echo(f'{metadata.program}/{metadata.transaction}')
+        typer.echo('\nInputs:')
+        print(
+            tabulate.tabulate(
+                [
+                    input.__dict__ for input in
+                    metadata.inputs
+                ],
+                headers='keys'
+            )
+        )   
 
-    typer.echo('\nOutputs:')
-    print(
-        tabulate.tabulate(
-            [
-                output.__dict__ for output in
-                records.outputs
-            ],
-            headers='keys'
+        typer.echo('\nOutputs:')
+        print(
+            tabulate.tabulate(
+                [
+                    output.__dict__ for output in
+                    metadata.outputs
+                ],
+                headers='keys'
+            )
         )
-    )
+    elif isinstance(metadata, MiError):
+        print(metadata)
 
-
-def print_program_meta(name: str, program: str):
-    endpoint = mi_endpoint_get(name)
-    
-    api = MiApi(endpoint.host, endpoint.port, endpoint.usr, endpoint.pwd)
-    records = api.program_meta(program)
-
-    typer.echo(f'{records.program} - {records.description}\n')
-    
-    print(
-        tabulate.tabulate(
-            [
-                {
-                    'transaction': transaction.transaction,
-                    'description': transaction.description
-                }
-                for transaction in
-                records.transactions
-            ],
-            headers='keys'
-        )
-    )
-
-
-def print_program_list(name: str):
-    endpoint = mi_endpoint_get(name)
-    
-    api = MiApi(endpoint.host, endpoint.port, endpoint.usr, endpoint.pwd)
-    records = api.program_list()
-
-    for row in records.records:
-        typer.echo(row)
-
-
-def print_endpoint_list():
-    print(
-        tabulate.tabulate(
-            [
-                {
-                    key: val for (key, val) in endpoint.__dict__.items()
-                    if key != 'pwd'
-                }
-                for endpoint
-                in mi_endpoint_list()
-            ],
-            headers='keys'
-        )
-    )
+    else:
+        raise Exception('Undefined return type, {type(metadata)}')
 
 if __name__ == '__main__':
     app()
\ No newline at end of file
